library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ov7670_udp_bridge is

  port (
    reset : in std_logic;
    
    pOV7670_PixelClk : in std_logic;
    pLineCounter     : in std_logic_vector(15 downto 0);

    pClk_50M  : in  std_logic;
    pMemAddr  : out std_logic_vector(9 downto 0);
    pMemData0 : in  std_logic_vector(31 downto 0);
    pMemData1 : in  std_logic_vector(31 downto 0);
    
    pO0Data    : out std_logic_vector(31 downto 0);
    pO0Request : out std_logic;
    pO0Ack     : in  std_logic;
    pO0Enable  : out std_logic;
    
    pSrcIPAddr  : in std_logic_vector(31 downto 0);
    pDestIPAddr : in std_logic_vector(31 downto 0);
    pSrcPort    : in std_logic_vector(15 downto 0);
    pDestPort   : in std_logic_vector(15 downto 0);
    
    debug : out std_logic_vector(7 downto 0)
    );
  
end ov7670_udp_bridge;

architecture RTL of ov7670_udp_bridge is

  component upl_tx
    port (
      pReset_n : in std_logic;
      pUPLGlobalClk : in std_logic;
      
      -- UPLに出力する
      pO0Data    : out std_logic_vector(31 downto 0 );
      pO0Request : out std_logic;
      pO0Ack     : in  std_logic;
      pO0En      : out std_logic;
      
      pMemAddr : out std_logic_vector(9 downto 0);
      pMemData : in  std_logic_vector(31 downto 0);

      pStart : in  std_logic;
      pBusy  : out std_logic;
      pDataLength : in std_logic_vector(31 downto 0);

      pSrcIPAddr   : in std_logic_vector(31 downto 0);
      pDestIPAddr  : in std_logic_vector(31 downto 0);
      pSrcPort     : in std_logic_vector(15 downto 0);
      pDestPort    : in std_logic_vector(15 downto 0);
      pLineCounter : in std_logic_vector(15 downto 0);

      pDebugA : out std_logic_vector( 7 downto 0 )
      );
  end component;

  signal reset_n : std_logic;
  
  signal line_counter_reg  : std_logic_vector(15 downto 0);

  signal cMemData   : std_logic_vector(31 downto 0);
  signal send_start : std_logic;
  signal send_busy  : std_logic;

  signal ov7670_send_line      : std_logic_vector(15 downto 0);
  signal ov7670_send_start     : std_logic;
  signal ov7670_send_start_reg : std_logic;
  signal ov7670_send_from      : std_logic;

  -- 640 pixel data (16bit) are read as 32bit, therefore the length is 320
  -- In addtion to, the line id is added at the first word.
  constant send_data_length : std_logic_vector(31 downto 0) := X"00000141";  -- 321
  
  attribute keep : string;
  attribute keep of send_busy : signal is "true";
  attribute keep of ov7670_send_from : signal is "true";

  signal sel_buffer  : std_logic;

begin  -- RTL

  reset_n <= not reset;
  
  process (pOV7670_PixelClk)
  begin  -- process
    if pOV7670_PixelClk'event and pOV7670_PixelClk = '1' then  -- rising clock edge
      if reset = '1' then
        line_counter_reg  <= (others => '0');
        ov7670_send_from  <= '0';
        ov7670_send_start <= '0';
        sel_buffer        <= '0';
      else
        line_counter_reg <= pLineCounter;
        if pLineCounter /= line_counter_reg then
          if pLineCounter = X"0000" then
            sel_buffer <= not sel_buffer;
          end if;
          if line_counter_reg(0) = sel_buffer then
            ov7670_send_start <= '1';
            ov7670_send_line  <= line_counter_reg;
            ov7670_send_from  <= sel_buffer;
          end if;
        elsif ov7670_send_start = '1' then
          ov7670_send_start <= '0';
        end if;
      end if;
    end if;
  end process;

  process (pClk_50M)
  begin  -- process
    if pClk_50M'event and pClk_50M = '1' then  -- rising clock edge
      if reset = '1' then
        send_start            <= '0';
        ov7670_send_start_reg <= '0';
      else
        ov7670_send_start_reg <= ov7670_send_start;
        if ov7670_send_start_reg = '0' and ov7670_send_start = '1' then
          send_start <= '1';
        elsif send_start = '1' then
          send_start <= '0';
        end if;
      end if;
    end if;
  end process;

  cMemData <= pMemData0 when ov7670_send_from = '0' else pMemData1;
  
  U_UPL_TX : upl_tx port map(
      pReset_n      => reset_n,
      pUPLGlobalClk => pClk_50M,

      -- UPLに出力する
      pO0Data    => pO0Data,
      pO0Request => pO0Request,
      pO0Ack     => pO0Ack,
      pO0En      => pO0Enable,

      pMemAddr => pMemAddr,
      pMemData => cMemData,

      pStart      => send_start,
      pBusy       => send_busy,
      pDataLength => send_data_length,

      pSrcIPAddr  => pSrcIPAddr,
      pDestIPAddr => pDestIPAddr,
      pSrcPort    => pSrcPort,
      pDestPort   => pDestPort,
      pLineCounter => ov7670_send_line,
      
      pDebugA => open
      );

end RTL;
